#ifndef __CSGJS_CONSTANTS__
#define __CSGJS_CONSTANTS__

typedef double csgjs_real;

//#define CSGJS_DEBUG
const csgjs_real EPS = .00001;
const csgjs_real NEG_EPS = -EPS;
const csgjs_real EPS_SQUARED = EPS*EPS;

#endif
